import json
import re
from pathlib import Path
from warnings import warn

import pytest

from tinyhtml5 import constants
from tinyhtml5.tokenizer import HTMLTokenizer


class TokenizerTestParser:
    def __init__(self, initial_state, last_start_tag):
        self._state = initial_state
        self._last_start_tag = last_start_tag

    def parse(self, stream):
        tokenizer = HTMLTokenizer(stream)
        self.output_tokens = []

        tokenizer.state = getattr(tokenizer, self._state)
        if self._last_start_tag is not None:
            tokenizer.current_token = {"type": "startTag", "name": self._last_start_tag}

        types = {
            value: key.lower() for key, value in constants.Token.__members__.items()}
        for token in tokenizer:
            getattr(self, f"process_{types[token['type']]}")(token)

        return self.output_tokens

    def process_doctype(self, token):
        self.output_tokens.append([
            "DOCTYPE", token["name"], token["publicId"],
            token["systemId"], token["correct"]])

    def process_start_tag(self, token):
        self.output_tokens.append([
            "StartTag", token["name"], token["data"], token["selfClosing"]])

    def process_empty_tag(self, token):
        if token["name"] not in constants.void_elements:
            self.output_tokens.append("ParseError")
        self.output_tokens.append(
            ["StartTag", token["name"], dict(token["data"][::-1])])

    def process_end_tag(self, token):
        self.output_tokens.append(["EndTag", token["name"], token["selfClosing"]])

    def process_comment(self, token):
        self.output_tokens.append(["Comment", token["data"]])

    def process_space_characters(self, token):
        self.output_tokens.append(["Character", token["data"]])
        self.process_space_characters = self.process_characters

    def process_characters(self, token):
        self.output_tokens.append(["Character", token["data"]])

    def process_eof(self, token):
        pass

    def process_parse_error(self, token):
        self.output_tokens.append(["ParseError", token["data"]])


def concatenate_character_tokens(tokens):
    output_tokens = []
    for token in tokens:
        last_is_character = (
            output_tokens and "ParseError" not in token + output_tokens[-1] and
            token[0] == output_tokens[-1][0] == "Character")
        if last_is_character:
            output_tokens[-1][1] += token[1]
        else:
            output_tokens.append(token)
    return output_tokens


def tokens_match(expected, received):
    """Test whether the test has passed or failed.

    If the ignoreErrorOrder flag is set we don’t test the relative
    positions of parse errors and non parse errors.

    """
    check_self_closing = any(
        token[0] == "StartTag" and len(token) == 4 or
        token[0] == "EndTag" and len(token) == 3
        for token in expected)
    if not check_self_closing:
        for token in received:
            if token[0] == "StartTag" or token[0] == "EndTag":
                token.pop()

    # Sort the tokens into two groups: non-parse errors and parse errors.
    tokens = {"expected": [[], []], "received": [[], []]}
    for token_type, token_list in zip(tokens, (expected, received)):
        for token in token_list:
            if token != "ParseError":
                tokens[token_type][0].append(token)
        tokens[token_type][0] = concatenate_character_tokens(tokens[token_type][0])
    return tokens["expected"] == tokens["received"]


def decode(input):
    """Decode \\uXXXX escapes.

    This decodes \\uXXXX escapes, possibly into non-BMP characters when
    two surrogate character escapes are adjacent to each other.

    """
    # This cannot be implemented using the unicode_escape codec because that
    # requires its input be ISO-8859-1, and we need arbitrary unicode as input.
    return re.compile(r"\\u([0-9A-Fa-f]{4})").sub(
        lambda match: chr(int(match.group(1), 16)), input)


def unescape(test):
    test["input"] = decode(test["input"])
    for token in test["output"]:
        token[1] = decode(token[1])
    return test


_tests = tuple(
    (f"{path.stem}-{i}", test)
    for path in (Path(__file__).parent / "tokenizer").glob("*.test")
    for i, test in enumerate(json.loads(path.read_text())["tests"]))
_xfails = (
    "domjs-4",
    "test2-0",
    "test3-280", "test3-283", "test3-284", "test3-286", "test3-287", "test3-289",
    "test3-292", "test3-293", "test3-295", "test3-296", "test3-298", "test3-310",
    "test3-718",
)


@pytest.mark.parametrize("id, test", _tests, ids=(id for id, _ in _tests))
def test_tokenizer(id, test):
    if "initialStates" not in test:
        test["initialStates"] = ["Data state"]
    if "doubleEscaped" in test:
        test = unescape(test)
    for initial_state in test.get("initialStates"):
        initial_state = re.compile(r"\W+(\w)").sub(
            lambda match: f"_{match.group(1)}", initial_state.lower())
        expected = test["output"]
        parser = TokenizerTestParser(initial_state, test.get("lastStartTag"))
        tokens = parser.parse(test["input"])
        received = [token[0] if token[0] == "ParseError" else token for token in tokens]
        error_message = (
            f"\nInitial state: {initial_state}",
            f"\nInput: {test['input']}",
            f"\nExpected: {expected!r}",
            f"\nReceived: {tokens!r}")
        match = tokens_match(expected, received)
        if id in _xfails:
            if not match:
                pytest.xfail()
        else:
            assert match, error_message
    else:
        if id in _xfails:  # pragma: no cover
            warn(f"{id} passes but is marked as xfail")



from collections import defaultdict
from os import linesep


class Data:
    def __init__(self, filename, new_test_heading="data", encoding="utf8"):
        if encoding:
            self.fd = open(filename, encoding=encoding, newline=linesep)
        else:
            self.fd = open(filename, mode="rb")
        self.encoding = encoding
        self.new_test_heading = new_test_heading

    def __iter__(self):
        data = defaultdict(lambda: None)
        key = None
        for line in self.fd:
            # Remove trailing newline.
            line = line[:-len(linesep)]
            if line.startswith("#" if self.encoding else b"#"):
                heading = line[1:].strip()
                if data and heading == self.new_test_heading:
                    # Remove empty line.
                    data[key] = data[key][:-1]
                    yield data
                    data = defaultdict(lambda: None)
                key = heading
                data[key] = "" if self.encoding else b""
            elif key is not None:
                if data[key]:
                    data[key] += "\n" if self.encoding else b"\n"
                data[key] += line
        if data:
            yield data



from pathlib import Path

import pytest

from tinyhtml5.inputstream import HTMLBinaryInputStream
from tinyhtml5.parser import HTMLParser

from . import Data

_tests = tuple(
    (f"{path.stem}-{i}", test)
    for path in (Path(__file__).parent / "encoding").glob("*.dat")
    for i, test in enumerate(Data(path, b"data", encoding=None)))


@pytest.mark.parametrize("id, test", _tests, ids=(id for id, _ in _tests))
def test_parser_encoding(id, test):
    parser = HTMLParser()
    assert parser.encoding is None
    parser.parse(test[b"data"])
    encoding = test[b"encoding"].lower().decode("ascii")
    error_message = (
        f"\nData: {test[b'data']!r}",
        f"\nExpected encoding: {encoding}",
        f"\nParser encoding: {parser.encoding}")
    assert encoding == parser.encoding, error_message


@pytest.mark.parametrize("id, test", _tests, ids=(id for id, _ in _tests))
def test_prescan_encoding(id, test):
    stream = HTMLBinaryInputStream(test[b"data"])

    # Very crude way to ignore irrelevant tests.
    if len(test[b"data"]) > stream.number_bytes_meta:
        return

    encoding = test[b"encoding"].lower().decode("ascii")
    error_message = (
        f"\nData: {test[b'data']!r}"
        f"\nExpected encoding: {encoding}"
        f"\nParser encoding: {stream.encoding[0].name}")
    assert encoding == stream.encoding[0].name, error_message



import os
from pathlib import Path
from warnings import warn
from xml.etree import ElementTree

import pytest

from tinyhtml5 import constants
from tinyhtml5.parser import HTMLParser

from . import Data


def convert(data, strip_chars):
    return (line[strip_chars:] if line.startswith("|") else line for line in data)


def serialize(item, indent=0, skip=False):
    if item.tag == ElementTree.Comment:
        yield f"{'  ' * indent}<!-- {item.text} -->"
        return
    elif item.tag == "<!DOCTYPE>":
        text = item.text
        if public := item.get("publicId"):
            public = f"'{public}'" if '"' in public else f'"{public}"'
            text += f" PUBLIC {public}" if item.text == "html" else f' {public} ""'
        elif system := item.get("systemId"):
            system = f'"{system}"'
            text += " SYSTEM" if item.text == "html" else ' ""'
            text += f" {system}"
        yield f"<!DOCTYPE {text}>"
        return
    if not skip:
        yield f"{'  ' * indent}<{item.tag}>"
        indent += 1
        for key, value in sorted(item.attrib.items()):
            yield f'{"  " * indent}{key}="{value}"'
    if item.text:
        yield f'{"  " * indent}"{item.text}"'
    for child in item:
        yield from serialize(child, indent)
        if child.tail:
            yield f'{"  " * indent}"{child.tail}"'


_tests = tuple(
    (f"{path.stem}-{i}", test)
    for path in (Path(__file__).parent / "tree-construction").glob("*.dat")
    for i, test in enumerate(Data(path)) if "script-on" not in test)
_xfails = (
    "adoption01-17",

    "blocks-12",

    "foreign-fragment-2", "foreign-fragment-3", "foreign-fragment-18",
    "foreign-fragment-19", "foreign-fragment-22", "foreign-fragment-23",
    "foreign-fragment-26", "foreign-fragment-27", "foreign-fragment-30",
    "foreign-fragment-31", "foreign-fragment-34", "foreign-fragment-35",
    "foreign-fragment-39", "foreign-fragment-41", "foreign-fragment-60",
    "foreign-fragment-61", "foreign-fragment-62", "foreign-fragment-63",
    "foreign-fragment-66",

    "isindex-0", "isindex-1", "isindex-2", "isindex-3",

    "namespace-sensitivity-0",

    "ruby-0", "ruby-1", "ruby-2", "ruby-3", "ruby-5", "ruby-7",
    "ruby-10", "ruby-12", "ruby-15", "ruby-17", "ruby-20",

    *[f"template-{i}" for i in range(7)],
    *[f"template-{i}" for i in range(8, 39)],
    *[f"template-{i}" for i in range(40, 111)],

    "tests2-6", "tests2-7",

    "tests8-5",

    "tests11-2", "tests11-4", "tests11-5", "tests11-6",

    "tests18-15",

    "tests19-7", "tests19-14", "tests19-17",

    "tests25-7",

    "tests26-16", "tests26-17", "tests26-18", "tests26-19",

    "webik02-14", "webik02-15", "webik02-16",

    # Introduced by the fork.

    # Doctype.
    "doctype01-26", "doctype01-28", "doctype01-29", "doctype01-31", "doctype01-32",
    "doctype01-33", "doctype01-34", "doctype01-35", "doctype01-36",
    "tests6-46",
    "quirks01-0", "quirks01-1", "quirks01-2", "quirks01-3",

    # <template>, <search>.
    "template-111", "search-element-0", "search-element-1",

    # Mismatched elements.
    "webkit02-11", "webkit02-15", "webkit02-16", "webkit02-17",

    # Attributes namespace.
    "webkit02-22",

    # <hr> in <options>.
    *[f"webkit02-{i}" for i in range(25, 35)],

    # Namespace in fragment.
    "foreign-fragment-59", "foreign-fragment-65",
)
if os.name == "nt":
    # Fail because of newline management on Windows.
    _xfails += (
        "domjs-unsafe-0",
        "pending-spec-changes-plain-text-unsafe-0",
        "plain-text-unsafe-0",
    )

@pytest.mark.parametrize("namespace", (False, True), ids=("nons", "ns"))
@pytest.mark.parametrize("id, test", _tests, ids=(id for id, _ in _tests))
def test_tree_construction(id, test, namespace):
    # TODO: Check error messages.
    parser = HTMLParser(namespace_html_elements=namespace)

    input = test["data"]
    container = test["document-fragment"]
    expected = "\n".join(convert(test.get("document", "").split("\n"), 2))

    if container:
        document = parser.parse_fragment(input, container=container)
    else:
        document = parser.parse(input, full_tree=True)

    output = "\n".join(convert(serialize(document, skip=True), 3))
    for prefix, value in constants.prefixes.items():
        output = output.replace(f"{{{prefix}}}", "" if value == "html" else f"{value} ")

    if id in _xfails:
        if expected == output:  # pragma: no cover
            warn(f"{id}{'-ns' if namespace else ''} passes but is marked as xfail")
        else:
            pytest.xfail()
    else:
        error_message = "".join(
            f"\n{line}\n" for line in (
                f"Input:\n{input}",
                f"Expected:\n{expected}",
                f"Received:\n{output}"))
        assert expected == output, error_message
